#!/usr/bin/env python
# encoding: utf-8
#
# help.py
#
# Created by José Sánchez-Gallego on 19 Mar 2017.


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import json

import click

__all__ = ('cheatSheet')


@click.command()
@click.pass_context
def cheatSheet(ctx):
    """Shows the help."""

    cmd = ctx.obj['cmd']

    for line in ctx.parent.get_help().splitlines():
        line = json.dumps(line).replace(';', '')
        cmd.writeToUsers('w', 'text={0}'.format(line))



#!/usr/bin/env python
# encoding: utf-8
#
# sms.py
#
# Created by John Donor on 1 Mar 2019.

from twilio.rest import Client
def sendSms(keyState, phoneNumbers=["+18177733196"]):
    # $TWILIO_ACCOUNT_SID and $TWILIO_AUTH_TOKEN must be set
    client = Client()

    severity = keyState.severity
    
    alertTxt = "{} alert on {}".format(severity, keyState.actorKey)

    for n in phoneNumbers:
        print("sms to {}".format(n))
        message = client.messages \
                        .create(
                            body=alertTxt,
                            from_='+19726456866',
                            to=n
                        )
        
        result = client.messages.get(message.sid).fetch()
        print("sms to {} was {}".format(n, result.status))



#!/usr/bin/env python
# encoding: utf-8
#
# dangerKey.py
#
# Created by John Donor on 10 April 2019

import re, time
from yaml import YAMLObject

from alertsActor import log


class diskCheck(YAMLObject):
    """evaluate a disk keyword
    """
    def __init__(self):
        pass

    def __call__(self, keyState):
        """The keyval is an enum ('Ok','Warning','Serious','Critical')
        and the amount of free space (GB)
        """
        keyval = keyState.keyword
        if (keyval[0]).upper() == 'OK':
            return "ok"
        elif (keyval[0]).upper() == 'WARNING':
            return "warn"
        elif (keyval[0]).upper() == 'SERIOUS':
            return "serious"
        elif (keyval[0]).upper() == 'CRITICAL':
            return "critical"
        else:
            return "info"


class doNothing(object):
    """camcheck alerts can't check themselves

       dummy class to facilitate that
    """
    def __init__(self):
        pass

    def __call__(self, keyState):
        return keyState.severity


class camCheck(YAMLObject):
    """evaluate a camCheck alert
    """
    def __init__(self):
        # NEVER GETS CALLED!!!! -_-
        pass

    def generateCamCheckAlert(self, key, severity):
        inst = key[:3]
        side = key[3]
        key = "camCheck." + key
        instruments = ["boss"]

        # most keywords will be SP[12][RB]
        # check if they are and assign appropriate instruments
        if inst in ["SP1", "SP2"]:
            instruments.append("boss.{}".format(inst))
            if side in ["R", "B"]:
                instruments.append("boss.{}.{}".format(inst, side))

        if severity in ["critical", "serious"]:
            selfClear = False
            addresses = self.emailAddresses
        else:
            selfClear = True
            addresses = None

        if key not in self.triggered:
            self.triggered.append(key)

        if key not in self.alertsActor.monitoring:
            dumbCheck = doNothing()
            self.alertsActor.addKey(key, severity=severity, checkAfter=120,
                                    selfClear=selfClear, checker=dumbCheck,
                                    keyword="'Reported by camCheck'",
                                    instruments=instruments, emailAddresses=addresses,
                                    emailDelay=0)
        if self.alertsActor.monitoring[key].active:
            self.alertsActor.monitoring[key].stampTime()
        else:
            self.alertsActor.monitoring[key].setActive(severity)

    def __call__(self, keyState):
        keyval = keyState.keyword
        if self.alertsActor is None:
            print("setting alertsActor for camCheck!!")
            self.alertsActor = keyState.alertsActorReference
            # do this only once hopefully
            for i in ["boss.SP1", "boss.SP2", "boss.SP1.R", "boss.SP2.R",
                      "boss.SP1.B", "boss.SP2.B"]:
                self.alertsActor.instrumentDown[i] = False
        # print("CAMCHECK, len {}, type {}, key: {}".format(len(keyval), type(keyval), keyval))
        log.info('CAMCHECK reported {}'.format(keyval))

        if type(keyval) == str:
            # could possibly try to fix this in hubModel casts, but easier here
            keyval = [keyval]
        if len(keyval) == 1 and keyval[0] == "None": # this is a bug somewhere upstream
            keyval = []
        for k in keyval:
            if re.search(r"SP[12][RB][0-3]?CCDTemp", k):
                self.generateCamCheckAlert(k, "critical")
            elif re.search(r"SP[12]SecondaryDewarPress", k):
                self.generateCamCheckAlert(k, "critical")
            elif re.search(r"SP[12](DAQ|Mech|Micro)NotTalking", k):
                self.generateCamCheckAlert(k, "critical")
            elif re.search(r"DACS_SET", k):
                self.generateCamCheckAlert(k, "critical")
            elif re.search(r"SP[12]LN2Fill", k):
                self.generateCamCheckAlert(k, "serious")
            elif re.search(r"SP[12](Exec|Phase)Boot", k):
                self.generateCamCheckAlert(k, "serious")
            else:
                self.generateCamCheckAlert(k, "warn")

        for k in self.triggered:
            if k.split(".")[-1] not in keyval:  # b/c we know its camCheck already
                self.alertsActor.monitoring[k].severity = "ok"
                # now it can check itself and find out its cool
                # and then decide to disappear if its acknowledged, etc etc
                self.alertsActor.monitoring[k].checkKey()
                self.triggered.remove(k)

        # never flag camCheck, always monitored keys
        return "ok"


class heartbeatCheck(YAMLObject):
    """check a heartbeat.
    """
    def __init__(self):
        pass

    def __call__(self, keyState):
        if time.time() - keyState.lastalive < keyState.checkAfter:
            return "ok"
        elif time.time() - keyState.lastalive > 5*keyState.checkAfter:
            return "critical"
        else:
            return keyState.defaultSeverity


class above(YAMLObject):
    """literally: is the value too high
    """
    def __init__(self):
        pass

    def __call__(self, keyState):
        if keyState.keyword > keyState.dangerVal:
            return keyState.defaultSeverity
        else:
            return "ok"


class below(YAMLObject):
    """literally: is the value too low
    """
    def __init__(self):
        pass

    def __call__(self, keyState):
        if keyState.keyword < keyState.dangerVal:
            return keyState.defaultSeverity
        else:
            return "ok"


class neq(YAMLObject):
    """literally: is the value too low
    """
    def __init__(self):
        pass

    def __call__(self, keyState):
        if keyState.keyword != keyState.dangerVal:
            return keyState.defaultSeverity
        else:
            return "ok"


class inList(YAMLObject):
    """is any value in the list "True", e.g. flagged
    """
    def __init__(self):
        pass

    def __call__(self, keyState):
        if [k for k in keyState.keyword if k]:
            return keyState.defaultSeverity
        else:
            return "ok"


class firstElem(YAMLObject):
    """is any value in the list "True", e.g. flagged
    """
    def __init__(self):
        pass

    def __call__(self, keyState):
        if keyState.keyword[0] == keyState.dangerVal:
            return keyState.defaultSeverity
        else:
            return "ok"


class default(object):
    """check equality to a dangerval
    """
    def __init__(self):
        pass

    def __call__(self, keyState):
        if keyState.keyword == keyState.dangerVal:
            return keyState.defaultSeverity
        else:
            return "ok"



#!/usr/bin/env python
# encoding: utf-8
#
# ping.py
#
# Created by José Sánchez-Gallego on 19 Mar 2017.


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import click

from alertsActor.cmds import alerts_context


__all__ = ('ping')


@click.command()
@alerts_context
def ping(actor, cmd, user):
    """Pings the actor."""

    # for k, a in actor.monitoring.items():
    #     print(k, a.keyword)

    cmd.setState(cmd.Done, "I'm not dead yet!")
    # print("monitoring heartbeats for :", actor.heartbeats.keys())

    return False



#!/usr/bin/env python
# encoding: utf-8
#
# version.py
#
# Created by José Sánchez-Gallego on 17 Feb 2017.


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import click
from alertsActor.cmds import alerts_context

from alertsActor import __version__


__all__ = ('version')


@click.command()
@alerts_context
def version(actor, cmd, user):
    """Returns the version."""

    cmd.writeToUsers('i', 'version="{0}"'.format(__version__))
    cmd.setState('done')

    return False




import click

from functools import update_wrapper


def alerts_context(f):
    # Unpacks the context and send the actor and command.
    @click.pass_context
    def new_func(ctx, *args, **kwargs):
        return ctx.invoke(f, ctx.obj['actor'], ctx.obj['cmd'], ctx.obj['user'],
                          *args, **kwargs)
    return update_wrapper(new_func, f)



#!/usr/bin/env python
# encoding: utf-8
#
# alerts_main.py
#


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import json
# import yaml
import sys
import os
import traceback

import time

from click.testing import CliRunner

from opscore.RO.StringUtil import strFromException
from opscore.RO.Comm.TwistedTimer import Timer
from opscore.utility.qstr import qstr
from twistedActor import BaseActor, CommandError, UserCmd

from alertsActor import __version__, alertActions
from alertsActor.cmds.cmd_parser import alerts_parser
from alertsActor import log

from alertsActor.rules import callbackWrapper, mail, dangerKey  # sms


class alertsActor(BaseActor):
    """the actor"""

    def __init__(self, config, **kwargs):

        self.cmdParser = alerts_parser
        self.config = config

        super(alertsActor, self).__init__(**kwargs)

        # a dictionary of actor keys we're watching
        self.monitoring = dict()

        # keep track of downed instruments
        # since its either up or down, make boolean
        # UP = False, DOWN = True
        self.instrumentDown = dict()

        # keep track of heartbeats
        self.heartbeats = dict()

        self.callbacks = callbackWrapper.wrapCallbacks(self, alertActions)

        self.connectHub('localhost', datamodel_casts=self.callbacks.datamodel_casts,
                                     datamodel_callbacks=self.callbacks.datamodel_callbacks)

        logfile = os.path.join(self.config["logging"]["logdir"],
                               time.strftime("%Y-%m-%dT%H:%M:%S", time.gmtime()))
        log.start_file_logger(logfile + ".log")

        log.info('starting alertsActor actor version={!r} in port={}'
                 .format(__version__, kwargs['userPort']))

    def addKey(self, key, severity, **kwargs):
        self.monitoring[key] = keyState(self, actorKey=key, severity=severity, **kwargs)

    @property
    def activeAlerts(self):
        active = []
        for k, a in self.monitoring.items():
            if a.active:
                active.append(a)

        return active

    @property
    def disabledAlerts(self):
        disabled = []
        for k, a in self.monitoring.items():
            if a.disabled:
                disabled.append(a)

        return disabled

    def broadcastActive(self):
        activeMessage = "activeAlerts{}".format(("=" if len(self.activeAlerts) else "")) +\
                       ", ".join(["{}".format(a.actorKey) for a in self.activeAlerts])
        self.writeToUsers("i", activeMessage)

    def broadcastDisabled(self):
        disabledMessage = "disabledAlertRules{}".format(("=" if len(self.disabledAlerts) else "")) +\
                       ", ".join(['"({}, {}, {})"'.format(a.actorKey, a.severity, a.disabledBy)
                                  for a in self.disabledAlerts])
        self.writeToUsers("i", disabledMessage)

    def broadcastAll(self):
        for a in self.activeAlerts:
            a.dispatchAlertMessage()

    def broadcastInstruments(self):
        instruments = list()
        for k, a in self.monitoring.items():
            if a.instruments is not None:
                for i in a.instruments:
                    if i not in instruments:
                        instruments.append(i)
        down = [i for i in instruments if self.instrumentDown[i]]
        self.writeToUsers("i", 'instrumentNames={}'.format(",".join(instruments)))
        self.writeToUsers("i", 'downInstruments={}'.format(",".join(down)))

    def reconnect(self):
        # convenience method to call occassionally
        # should be able to recover from a hub disconnect
        if self.hub is None:
            print("no hub connection, reconnecting!")
            self.connectHub('localhost', datamodel_casts=self.callbacks.datamodel_casts,
                                         datamodel_callbacks=self.callbacks.datamodel_callbacks)
        elif self.hub.didFail:
            print("no hub connection, reconnecting!")
            self.connectHub('localhost', datamodel_casts=self.callbacks.datamodel_casts,
                                         datamodel_callbacks=self.callbacks.datamodel_callbacks)

    @property
    def hubModel(self):
        # keeps a running data model of keywords coming from the hub
        # allows callbacks on updates

        # this may need to be more careful... test!
        self.reconnect()

        return self.hub.datamodel

    def parseAndDispatchCmd(self, cmd):
        """Dispatch the user command. Stolen from BMO."""

        def test_cmd(args):
            # print("t ", args)
            result = CliRunner().invoke(alerts_parser, args)
            if result.exit_code > 0:
                # If code > 0, there was an error. We fail the command and inform the users.
                textMsg = result.output
                for line in textMsg.splitlines():
                    line = json.dumps(line).replace(';', '')
                    cmd.writeToUsers('w', 'text={0}'.format(line))
                cmd.setState(cmd.Failed)
                return False
            else:
                if '--help' in args:
                    # If help was in the args, we just want to print the usage to the users.
                    textMsg = result.output
                    for line in textMsg.splitlines():
                        line = json.dumps(line).replace(';', '')
                        cmd.writeToUsers('w', 'text={0}'.format(line))
                    cmd.setState(cmd.Done)
                    return False

                return True

        if not cmd.cmdBody:
            # echo to show alive
            self.writeToOneUser(":", "", cmd=cmd)
            return

        cmd.setState(cmd.Running)

        try:
            # stui wants to pass keyword args, parse doesn't support that
            # this is definitely bad form, but we're trying not to change
            # stui yet
            # click is case sensetive and requires lower-case -_-
            # except keywords are often camelCase. Ugh.... Starting to dislike click
            args = [a.lower() if not "=" in a else a for a in cmd.cmdBody.split()]
            args = [a.split("=")[-1] for a in args]
            # current tron config passes user, we aren't handling that yet
            # print("cmd", cmd.__dict__)
            # print(args)
            if "." in args[0]:
                user = args[0]
                cmd.cmdID = int(args[1])
                temp_args = args[2:]
            else:
                # user = "?.?"
                user = f"{cmd.userID}.{cmd.cmdID}"
                temp_args = args
            args = temp_args
            # user = f"{cmd.userID}.{cmd.cmdID}"
            log.info('{} issued {}'.format(user, args))
            result = test_cmd(args)
            if result is False:
                return
            # cmd.cmdBody = " ".join(temp_args)
            alerts_parser(args, obj=dict(actor=self, cmd=cmd, user=user))
        except CommandError as ee:
            log.warning("command {} failed with {}".format(cmd.cmdStr, strFromException(ee)))
            cmd.setState('failed', textMsg=strFromException(ee))
            return
        except Exception as ee:
            sys.stderr.write('command {0!r} failed\n'.format(cmd.cmdStr))
            traceback.print_exc(file=sys.stderr)
            textMsg = strFromException(ee)
            hubMsg = 'Exception={0}'.format(ee.__class__.__name__)
            log.warning("command {} failed with {}".format(cmd.cmdStr, strFromException(ee.__class__.__name__)))
            cmd.setState("failed", textMsg=textMsg, hubMsg=hubMsg)
        except BaseException:
            # This catches the SystemExit that Click insists in returning.
            pass


def ack(acknowledged):
    if acknowledged:
        return "ack"
    else:
        return "noack"


def enabled(disabled):
    if disabled:
        return "disabled"
    else:
        return "enabled"


def parseKey(keyVal):
    """readable keyVal"""
    if type(keyVal) is list:
        return "[{}]".format("".join(["{}.".format(i) for i in keyVal]))
    else:
        return str(keyVal)


class keyState(object):
    '''Keep track of the state of each actor.key'''

    def __init__(self, alertsActor, actorKey='oop.forgot', keyword="",
                 emailAddresses=None, **kwargs):
        self.alertsActorReference = alertsActor
        self.triggeredTime = None
        self.actorKey = actorKey
        self.keyword = keyword
        self.lastalive = time.time()  # updated for heartbeats
        self.active = False
        self.disabled = False
        self.disabledBy = -1
        self.severity = "ok"
        self.acknowledged = False
        self.acknowledgeMsg = ""
        self.acknowledger = -1
        self.checkMe = Timer()
        self.emailTimer = Timer()
        self.emailAddresses = emailAddresses
        self.emailSent = False
        # self.smtpclient = "localhost:1025"
        self.smtpclient = alertsActor.config["email"]["mailClient"]

        # kwargs, second argument is the default
        self.defaultSeverity = kwargs.get("severity", "info")
        self.dangerVal = kwargs.get("dangerVal", None)
        self.selfClear = kwargs.get("selfClear", False)
        self.instruments = kwargs.get("instruments", None)
        self.checkAfter = kwargs.get("checkAfter", 120)
        self.checker = kwargs.get("checker", dangerKey.default())
        self.emailDelay = kwargs.get("emailDelay", self.checkAfter)

        assert self.severity in ['ok', 'info', 'apogeediskwarn', 'warn', 'serious', 'critical'], "severity level not allowed"

    def keywordFmt(self):
        if "heartbeat" in self.actorKey:
            instring = "at {time}; last seen {diff} sec ago".format(time=self.triggeredTime,
                                                                    diff=int(time.time()-self.lastalive))
        elif "stale" in self.actorKey:
            instring = "at {time}; no change for {diff} sec, with {keyword}".format(time=self.triggeredTime,
                                                                    diff=int(time.time()-self.lastalive),
                                                                    keyword=parseKey(self.keyword))
        else:
            instring = 'at {time} UT found {keyword}'.format(keyword=parseKey(self.keyword), time=self.triggeredTime)
        return qstr(instring)

    @property
    def msg(self):
        return "alert={actorKey}, {severity}, {value}, {enable}, {acknowledged}, {acknowledger}".format(
               actorKey=self.actorKey, value=self.keywordFmt(),
               severity=self.severity, enable=enabled(self.disabled), acknowledged=ack(self.acknowledged),
               acknowledger=self.acknowledger)

    @property
    def instDown(self):
        if self.instruments is None:
            return False

        for i in self.instruments:
            if self.alertsActorReference.instrumentDown[i]:
                return True
        else:
            return False

    def stampTime(self):
        self.triggeredTime = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())

    def setActive(self, severity=None):
        # something cause a problem, do stuff
        self.acknowledged = False  # clear anything from old alert
        self.active = True
        if severity is None:
            self.severity = self.defaultSeverity
        else:
            self.severity = severity
        self.stampTime()
        self.checkMe.start(self.checkAfter, self.reevaluate)

        if self.instDown:
            self.disable(0)
            log.info("NO ALERT: {} instrument down, no alert!!".format(self.actorKey))
            return None

        self.dispatchAlertMessage()
        # give the alert a chance to clear before emailing everyone
        self.emailTimer.start(self.emailDelay, self.sendEmail)

    def clear(self):
        # everything good, back to normal
        self.active = False
        self.checkMe.cancel()
        self.checkMe = Timer()
        self.emailTimer.cancel()
        self.emailTimer = Timer()
        self.severity = 'ok'
        self.triggeredTime = None
        self.emailSent = False

        self.disabledBy = -1
        self.acknowledger = -1
        self.acknowledgeMsg = ""

        self.alertsActorReference.broadcastActive()
        self.alertsActorReference.broadcastDisabled()
        self.alertsActorReference.broadcastAll()

    def acknowledge(self, msg=None, acknowledgedBy=None, unack=False):
        if msg is not None:
            self.acknowledgeMsg += msg + ";" # so we can add many... I guess?

        if acknowledgedBy is not None:
            self.acknowledger = acknowledgedBy

        if unack:
            self.acknowledged = False
        else:
            self.acknowledged = True

        self.dispatchAlertMessage()

    def reevaluate(self):
        # at some point this will presumably raise alert level?
        # possibly send email? Or only send email for critical? Or...
        if self.alertsActorReference.hub.didFail:
            self.alertsActorReference.reconnect()
        check = self.checkKey()
        print("evaluated {} found {} default {}".format(self.actorKey, check, self.defaultSeverity))
        if check != self.severity:
            # something changed, treat it like new alert
            self.severity = check
            self.acknowledge(acknowledgedBy=0, unack=True)
            self.dispatchAlertMessage()
        if not self.active:
            return

        if self.acknowledged:
            # keep checking myself until I go away
            self.checkMe.start(self.checkAfter, self.reevaluate)
            return

        if self.disabled:
            self.checkMe.start(self.checkAfter, self.reevaluate)
            return

        self.dispatchAlertMessage()
        self.checkMe.start(self.checkAfter, self.reevaluate)

    def sendEmail(self):
        # notify over email
        if self.emailAddresses is None or self.disabled:
            return

        if self.emailSent:
            # I don't think this should happen but it seems to...
            log.info("Tried to send extra email for {}".format(self.actorKey))
            return

        if not self.active:
            log.warn("Tried to send email after clear for {}".format(self.actorKey))
            return

        mail.sendEmail(self, self.smtpclient)
        # and sms?
        # sms.sendSms(self)  # just a reminder for later , phoneNumbers=["+18177733196"])
        self.emailSent = True

    def dispatchAlertMessage(self):
        # write an alert to users

        if self.severity == 'critical':
            broadcastSeverity = 'e'
        elif self.severity == 'warn' or self.severity == 'serious':
            broadcastSeverity = 'w'
        else:
            broadcastSeverity = 'i'

        self.alertsActorReference.writeToUsers(broadcastSeverity, self.msg)

        log.info("ALERT! " + self.msg)

    def checkKey(self):
        # check key, should be called when keyword changes

        check = self.checker(self)

        if check == "ok":
            if self.selfClear:
                self.clear()
            elif self.acknowledged:
                self.clear()
            else:
                # self.severity = "ok"
                return self.severity
        else:
            if not self.active:
                self.setActive(check)

        return check

    def disable(self, disabledBy):
        self.disabled = True
        self.disabledBy = disabledBy

        # self.active = False

        # probably do some more stuff

    def enable(self):
        self.disabled = False
        self.checkKey()



#!/usr/bin/env python
# encoding: utf-8
#
# instrumentState.py
#


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import click

from alertsActor.cmds import alerts_context

__all__ = ('instrumentState')


@click.command()
@click.argument('instrument', nargs=1, default=None, required=True)
@click.argument('state', nargs=1, default='up', type=click.Choice(['up', 'down']))
@alerts_context
def instrumentState(actor, cmd, user, instrument=None, state="up"):
    """set the state of an instrument"""

    if state == "down":
        actor.instrumentDown[instrument] = True
        for a in actor.activeAlerts:
            if instrument in a.instruments:
                a.disable(user)
    else:
        actor.instrumentDown[instrument] = False
        for a in actor.disabledAlerts:
            if not a.instDown:
                a.enable()

    actor.broadcastActive()
    actor.broadcastDisabled()
    actor.broadcastAll()
    actor.broadcastInstruments()

    cmd.setState(cmd.Done, '{} set to {}'.format(instrument, state))

    return False



#!/usr/bin/env python
# encoding: utf-8
#
# enable.py



from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import click

from alertsActor.cmds import alerts_context

__all__ = ('enable')


@click.command()
@click.argument('alertkey', nargs=1, default=None, required=True)
@click.argument('severity', nargs=1, default='info', 
                type=click.Choice(['ok', 'info', 'apogeediskwarn','warn', 'serious', 'critical']))
@alerts_context
def enable(actor, cmd, user, alertkey=None, severity='info'):
    """enable an alert"""

    # if isinstance(alertkey, unicode):
    #     alertkey = str(alertkey)  # .decode("utf-8")

    keyword = actor.monitoring[alertkey]

    keyword.enable()

    actor.broadcastActive()
    actor.broadcastDisabled()
    actor.broadcastAll()

    cmd.setState(cmd.Done, 'enabled')

    return False



#!/usr/bin/env python
# encoding: utf-8
#
# status.py


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os
import yaml

import click

from alertsActor.cmds import alerts_context

__all__ = ('status')


@click.command()
@alerts_context
def status(actor, cmd, user):
    """returns actor status"""

    # print("current model: ")
    # for k, v in actor.hubModel.items():
    #     print(k, v)
    #     print("\n \n")

    actor.broadcastActive()
    actor.broadcastDisabled()
    actor.broadcastAll()
    actor.broadcastInstruments()

    cmd.setState(cmd.Done, "Now you know all I know")

    try:
        dontClutterSDSSUser = os.getlogin()
    except OSError:
        # when run as a daemon, getlogin will fail
        dontClutterSDSSUser = "sdss5"
    if "sdss" in dontClutterSDSSUser:
        with open("/data/logs/actors/alerts/alerts.hubModel.yaml", "w") as dump:
            print(yaml.dump(actor.hubModel), file=dump)

    # otherwise we can clutter the home directory a bit
    # write a log of the hubModel for posterity
    else:
        with open("alerts.hubModel.yaml", "w") as dump:
            print(yaml.dump(actor.hubModel), file=dump)


    return False



#!/usr/bin/env python
# encoding: utf-8
#
# cmds.py
#
# Created by José Sánchez-Gallego on 17 Feb 2017.


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import click

from alertsActor.cmds.cheatSheet import cheatSheet
from alertsActor.cmds.ping import ping
from alertsActor.cmds.version import version
from alertsActor.cmds.status import status
from alertsActor.cmds.acknowledge import acknowledge
from alertsActor.cmds.unacknowledge import unacknowledge
from alertsActor.cmds.enable import enable
from alertsActor.cmds.disable import disable
from alertsActor.cmds.instrumentState import instrumentState


__all__ = ('alerts_parser')


@click.group()
@click.pass_context
def alerts_parser(ctx):
    pass


alerts_parser.add_command(cheatSheet)
alerts_parser.add_command(ping)
alerts_parser.add_command(version)
alerts_parser.add_command(status)
alerts_parser.add_command(acknowledge)
alerts_parser.add_command(unacknowledge)
alerts_parser.add_command(enable)
alerts_parser.add_command(disable)
alerts_parser.add_command(instrumentState)


if __name__ == '__main__':
    alerts_parser()



# encoding: utf-8
#
# main.py


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from pytest import mark

from alertsActor.main import math


class TestMath(object):
    """Tests for the ``math`` function in main.py."""

    @mark.parametrize(('arg1', 'arg2', 'operator', 'result'),
                      [(1, 2, '+', 3), (2, 2, '-', 0), (3, 5, '*', 15), (10, 2, '/', 5)])
    def test_math(self, arg1, arg2, operator, result):

        assert math(arg1, arg2, arith_operator=operator) == result



#!/usr/bin/env python
# encoding: utf-8
#
# file.py
#
# Created by José Sánchez-Gallego on 17 Sep 2017.


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import logging

from twisted.internet import reactor

from alertsActor import config, __version__
# from alertsActor.logger import log
from alertsActor.alerts_main import alertsActor

import click



@click.command()
@click.option('-d', '--debug', is_flag=True, show_default=True,
              help='Debug mode.')
def alerts_cmd(debug=False):

    port = config['tron']['port']

    if debug:
        print("debug not treated specially at the moment. TODO?")
        # log.sh.setLevel(logging.DEBUG)
        # log.debug('alertsActor started in debug mode.')

    alertsActor(config, userPort=port, version=__version__)

    reactor.run()


if __name__ == '__main__':
    alerts_cmd()



#!/usr/bin/env python
# encoding: utf-8
#
# callbacks.py
#
# Created by John Donor on 9 Mar 2018.

import time
from opscore.RO.Comm.TwistedTimer import Timer

from alertsActor import log


class wrapCallbacks(object):
    """Pass in keywords read from a file
       along with their alert type.
    """

    def __init__(self, alertsActor, keywords):
        # keywords read from config
        self.alertsActor = alertsActor
        self.datamodel_casts = dict()
        self.datamodel_callbacks = dict()

        for key, actions in keywords.items():
            self.datamodel_casts[key] = actions['cast']

            initKeys = ('cast', 'severity', 'heartbeat', 'stale')

            otherArgs = {k: actions[k] for k in actions.keys() if k not in initKeys}

            if 'heartbeat' in actions.keys():
                alertKey = actions['heartbeat'] + '.heartbeat'
                self.alertsActor.addKey(alertKey,
                                        severity=actions['severity'],
                                        **otherArgs)
                callback = self.pulse(alertKey=alertKey,
                                      checkAfter=actions['checkAfter'])
                self.datamodel_callbacks[key] = callback

            elif 'stale' in actions.keys():
                staleKey = key + ".stale"
                self.alertsActor.addKey(staleKey, severity=actions['severity'],
                                        checkAfter=actions['stale'],
                                        selfClear=True,
                                        checker=staleCheck,
                                        instruments=actions['instruments'],
                                        emailAddresses=actions['emailAddresses'],
                                        emailDelay=actions.get('emailDelay', 120))

                alertKey = key
                self.alertsActor.addKey(alertKey,
                                        severity=actions['severity'],
                                        **otherArgs)

                requireChange = actions.get("requireChange", False)

                callback = self.updateKeyStale(key,
                                               checkAfter=actions['stale'],
                                               requireChange=requireChange)
                self.datamodel_callbacks[key] = callback

            else:
                alertKey = key
                self.alertsActor.addKey(alertKey,
                                        severity=actions['severity'],
                                        **otherArgs)
                callback = self.updateKey(key)
                self.datamodel_callbacks[key] = callback

            if "instruments" in actions.keys():
                for i in actions["instruments"]:
                    alertsActor.instrumentDown[i] = False

    def pulse(self, alertKey='NOT_SPECIFIED', checkAfter=30):
        """Update the heartbeat for a specified actor.
           The timer is restarted everytime its called, if it isn't restarted
           in time, a "dead actor" alert is raised
        """
        if alertKey not in self.alertsActor.heartbeats.keys():
            self.alertsActor.heartbeats[alertKey] = Timer()

        deadCallback = self.itsDeadJim(alertKey=alertKey)

        def startTime(newKeyval):
            # called as callback, so the updated key is passed by default
            # print('pulse: ', alertKey, newKeyval[0])
            log.info('{}: the actor said {}'.format(alertKey, newKeyval))
            # don't remember why we want the first element of the list here
            # but it shouldn't matter for heartbeat monitoring
            self.alertsActor.monitoring[alertKey].keyword = newKeyval[0]
            self.alertsActor.monitoring[alertKey].lastalive = time.time()
            self.alertsActor.heartbeats[alertKey].start(checkAfter, deadCallback)
            self.alertsActor.monitoring[alertKey].checkKey()

        startTime(["init"])  # call now so it raises error?

        return startTime

    def updateKeyStale(self, actorKey, checkAfter=30, requireChange=False):
        """Update the create a "stale" heartbeat for a specified actorKey.
           The timer is restarted everytime its called, if it isn't restarted
           in time, a "stale" alert is raised
        """
        staleKey = actorKey + ".stale"
        if staleKey not in self.alertsActor.heartbeats.keys():
            self.alertsActor.heartbeats[staleKey] = Timer()

        deadCallback = self.itsDeadJim(alertKey=staleKey)

        def check(newKeyval, init=False):
            if len(newKeyval) == 1:
                newKeyval = newKeyval[0]
            if init:
                self.alertsActor.monitoring[staleKey].keyword = 0
            if requireChange:
                # val has to change to reset timer, e.g. BPR
                # definitely don't want to do this for string keys
                if abs(self.alertsActor.monitoring[staleKey].keyword - newKeyval) < 0.1:
                    return None
            # also do the heartbeat
            self.alertsActor.monitoring[staleKey].keyword = newKeyval
            self.alertsActor.monitoring[staleKey].lastalive = time.time()
            self.alertsActor.heartbeats[staleKey].start(checkAfter, deadCallback)
            self.alertsActor.monitoring[staleKey].checkKey()

            if init:
                return None

            log.info('{}: the actor said {}'.format(actorKey, newKeyval))
            # called as callback, so the updated key is passed by default
            self.alertsActor.monitoring[actorKey].keyword = newKeyval
            self.alertsActor.monitoring[actorKey].checkKey()

        check([-9999], init=True)

        return check

    def updateKey(self, actorKey):
        def check(newKeyval):
            # print("---------------------------")
            # print(newKeyval)
            log.info('{}: the actor said {}'.format(actorKey, newKeyval))
            if len(newKeyval) == 1:
                newKeyval = newKeyval[0]
            # print("{} {} {}".format(actorKey, newKeyval, type(newKeyval)))
            # called as callback, so the updated key is passed by default
            self.alertsActor.monitoring[actorKey].keyword = newKeyval
            self.alertsActor.monitoring[actorKey].checkKey()

        return check

    def itsDeadJim(self, alertKey='NOT_SPECIFIED'):
        # some actor stopped outputting a keyword, raise the alarm!
        def setActive():
            self.alertsActor.monitoring[alertKey].setActive()

        return setActive


def staleCheck(keyState):
    """basically same as a heartbeat
    """
    if time.time() - keyState.lastalive < keyState.checkAfter:
        return "ok"
    elif time.time() - keyState.lastalive > 5*keyState.checkAfter:
        return "critical"
    else:
        return keyState.defaultSeverity




import os
import yaml

# Inits the logging system. Only shell logging, and exception and warning catching.
# File logging can be started by calling log.start_file_logger(name).
from .utils import get_logger

# Monkeypatches formatwarning and error handling

import click
import warnings


def warning_on_one_line(message, category, filename, lineno, file=None, line=None):

    basename = os.path.basename(filename)
    category_colour = click.style('[{}]'.format(category.__name__), fg='yellow')

    return '{}: {} ({}:{})\n'.format(category_colour, message, basename, lineno)


warnings.formatwarning = warning_on_one_line

warnings.filterwarnings(
    'ignore', 'Matplotlib is building the font cache using fc-list. This may take a moment.')


log = get_logger('alerts')


# Loads config

if os.path.isfile(os.path.expanduser('~/.alertsConfig.yml')):
    configFile = os.path.expanduser('~/.alertsConfig.yml')
else:
    print("Local config not found! Using default")
    configFile = os.path.join(os.path.dirname(__file__), 'etc/alerts.cfg')

if os.path.isfile(os.path.expanduser("~/.alertActions.yml")):
    actionsFile = os.path.expanduser("~/.alertActions.yml")
else:
    print("Local alert actions not found! Using default")
    actionsFile = os.path.join(os.path.dirname(__file__), 'etc/alertActions.yml')

try:
    config = yaml.load(open(configFile), Loader=yaml.FullLoader)

    alertActions = yaml.load(open(actionsFile), Loader=yaml.UnsafeLoader)
except AttributeError:
    # using pyyaml < 5, enforce old behavior
    config = yaml.load(open(configFile))

    alertActions = yaml.load(open(actionsFile))

__version__ = '2.2.2'



#!/usr/bin/env python
# encoding: utf-8
#
# unacknowledge.py
#


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import click

from alertsActor.cmds import alerts_context

__all__ = ('unacknowledge')


@click.command()
@click.argument('alertkey', nargs=1, default=None, required=True)
@click.argument('severity', nargs=1, default='info',
                type=click.Choice(['ok', 'info', 'apogeediskwarn', 'warn', 'serious', 'critical']))
@click.option('-m', '--message', multiple=True, default=None, help='a short message to hang on to')
@alerts_context
def unacknowledge(actor, cmd, user, alertkey=None, severity='info', message=None):
    """unacknowledge an alert"""

    # if isinstance(alertkey, unicode):
    #     alertkey = str(alertkey)  # .decode("utf-8")

    keyword = actor.monitoring[alertkey]

    if keyword.severity != severity:
        cmd.setState(cmd.Failed, "Severity does not match alert severity")
        return None

    # it seems messages can't be passed right. geez..
    if len(message) > 0:
        msg = "".join(("{} ".format(m) for m in message))
    else:
        msg = None

    keyword.acknowledge(msg=msg, acknowledgedBy=user, unack=True)
    cmd.setState(cmd.Done, 'unacknowledged')

    return False



import time
import os

import smtplib
from email.mime.text import MIMEText

from alertsActor import log


def sendEmail(keyState, mailClient):
    """Send an email.

    Input:
    -keyState: keyState object from alerts_main
    -mailClient: str, the mail server to connect to
    """

    # Adds the time to the text
    text = keyState.msg + '\n\nEmail sent on ' + time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime()) + '\n'

    severity = keyState.severity

    recipients = keyState.emailAddresses

    sender = recipients[0]

    observatory = os.getenv("OBSERVATORY")

    subject = "{} {} alert on {}".format(observatory, severity, keyState.actorKey)
    msg = MIMEText("{}\n\n{}".format(subject, text))

    msg['Subject'] = subject

    msg['From'] = "%sAlerts (SDSS-V %s Alerts)" % (2 * (severity.capitalize(), ))

    msg['Reply-to'] = sender

    msg['To'] = ', '.join(recipients)

    # Send mail through the mail server
    try:
        s = smtplib.SMTP()
        s.connect(mailClient)
        # Send the email - real from, real to, extra headers and content ...
        s.sendmail(sender, recipients, msg.as_string())
        log.info("Sent email for %s %s" % (keyState.severity, keyState.actorKey))
    except Exception as e:
        log.warn("Sending email warning: %s %s" % (msg, e))
    finally:
        s.close()
